from __future__ import annotations

import io
from collections.abc import Callable


class Flags:
    def __init__(self, value: int = 0, length: int = 32) -> None:
        self.value = value & ((1 << length) - 1)
        self.length = length

    def has(self, position: int) -> bool:
        return bool(self.value & (1 << position))

    def get(self, position: int) -> bool:
        return bool(self.value & (1 << position))

    def if_set[T](self, position: int, callback: Callable[[], T]) -> T | None:
        if self.has(position):
            return callback()
        return None

    def set(self, position: int, value: bool = True) -> Flags:
        if value:
            self.value |= 1 << position
        else:
            self.value &= ~(1 << position)
        return self

    def unset(self, position: int) -> Flags:
        self.value &= ~(1 << position)
        return self

    def write(self, writer: ByteWriter) -> ByteWriter:
        bits = self.value.to_bytes((self.length + 7) // 8, "big")
        writer.write(bits)
        return writer

    @classmethod
    def read(cls, reader: ByteReader, length: int) -> Flags:
        bits = int.from_bytes(reader.read((length + 7) // 8), "big")
        return Flags(bits, length)

    def __or__(self, other: Flags) -> Flags:
        return Flags(self.value | other.value, max(self.length, other.length))

    def __and__(self, other: Flags) -> Flags:
        return Flags(self.value & other.value, max(self.length, other.length))

    def __xor__(self, other: Flags) -> Flags:
        return Flags(self.value ^ other.value, max(self.length, other.length))

    def __invert__(self) -> Flags:
        return Flags(~self.value, self.length)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Flags):
            return False
        return self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)


class ByteWriter:
    def __init__(self, init: bytes | None = None) -> None:
        self.stream = io.BytesIO(init or b"")
        self.finished = False

    def write(self, data: bytes) -> ByteWriter:
        if self.finished:
            raise ValueError("Writer already finished")
        self.stream.write(data)
        return self

    def write_boolean(self, value: bool) -> ByteWriter:
        self.write(value.to_bytes(1, "big"))
        return self

    def write_big_int(self, value: int) -> ByteWriter:
        self.write(value.to_bytes(8, "big"))
        return self

    def write_int(self, value: int) -> ByteWriter:
        self.write(value.to_bytes(4, "big"))
        return self

    def write_short(self, value: int) -> ByteWriter:
        self.write(value.to_bytes(2, "big"))
        return self

    def write_byte(self, value: int) -> ByteWriter:
        self.write(value.to_bytes(1, "big"))
        return self

    def write_byte_array(self, value: bytes) -> ByteWriter:
        if len(value) > 0xFFFFFFFF:
            raise ValueError("Byte array too large")
        self.write_int(len(value))
        self.write(value)
        return self

    def write_string(self, value: str) -> ByteWriter:
        self.write_byte_array(value.encode("utf-8"))
        return self

    def write_flags(self, flags: Flags) -> ByteWriter:
        flags.write(self)
        return self

    def finish(self) -> bytes:
        if self.finished:
            raise ValueError("Writer already finished")
        self.finished = True
        return self.stream.getvalue()


class ByteReader:
    def __init__(self, buffer: bytes) -> None:
        self.stream = io.BytesIO(buffer)
        self.is_reading = False
        self.is_finished = False

    def __enter__(self) -> ByteReader:
        if self.is_reading:
            raise ValueError("Reader already reading")
        if self.is_finished:
            raise ValueError("Reader already finished")
        self.is_reading = True
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        if self.stream.read(1):
            raise ValueError("Reader not fully consumed")
        self.is_reading = False
        self.is_finished = True

    def read(self, size: int) -> bytes:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        if size < 0:
            raise ValueError("Size must be positive")
        return self.stream.read(size)

    def read_boolean(self) -> bool:
        return bool(int.from_bytes(self.read(1), "big"))

    def read_big_int(self) -> int:
        return int.from_bytes(self.read(8), "big")

    def read_int(self) -> int:
        return int.from_bytes(self.read(4), "big")

    def read_short(self) -> int:
        return int.from_bytes(self.read(2), "big")

    def read_byte(self) -> int:
        return int.from_bytes(self.read(1), "big")

    def read_byte_array(self) -> bytes:
        length = self.read_int()
        return self.read(length)

    def read_string(self) -> str:
        return self.read_byte_array().decode("utf-8")

    def read_flags(self, length: int) -> Flags:
        return Flags.read(self, length)



from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.table import TableType
from omu.network.packet import PacketType

from .plugin import PluginPackageInfo

PLUGIN_EXTENSION_TYPE = ExtensionType(
    "plugin",
    lambda client: PluginExtension(client),
    lambda: [],
)


class PluginExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return PLUGIN_EXTENSION_TYPE

    def __init__(self, client: Client):
        self.client = client
        self.plugins: dict[str, str | None] = {}

        self.client.network.register_packet(
            PLUGIN_REQUIRE_PACKET,
        )
        self.client.network.add_task(self.on_task)

    async def on_task(self):
        await self.client.send(PLUGIN_REQUIRE_PACKET, self.plugins)

    def require(self, plugins: dict[str, str | None]):
        if self.client.running:
            raise RuntimeError("Cannot require plugins after client has started")
        self.plugins.update(plugins)


PLUGIN_REQUIRE_PACKET = PacketType[dict[str, str | None]].create_json(
    PLUGIN_EXTENSION_TYPE,
    "require",
)
PLUGIN_ALLOWED_PACKAGE_TABLE = TableType.create_model(
    PLUGIN_EXTENSION_TYPE,
    "allowed_package",
    PluginPackageInfo,
)



from __future__ import annotations

import abc
import json
from collections.abc import Callable, Mapping
from typing import Protocol


class SerializeError(Exception):
    pass


class Serializable[T, D](Protocol):
    @abc.abstractmethod
    def serialize(self, item: T) -> D: ...

    @abc.abstractmethod
    def deserialize(self, item: D) -> T: ...


class JsonSerializable[T, D](Protocol):
    def to_json(self) -> D: ...

    @classmethod
    def from_json(cls, json: D) -> T: ...


class Serializer[T, D](Serializable[T, D]):
    def __init__(self, serialize: Callable[[T], D], deserialize: Callable[[D], T]):
        self._serialize = serialize
        self._deserialize = deserialize

    def serialize(self, item: T) -> D:
        return self._serialize(item)

    def deserialize(self, item: D) -> T:
        return self._deserialize(item)

    @classmethod
    def of[_T, _D](cls, serializer: Serializable[_T, _D]) -> Serializer[_T, _D]:
        return Serializer(serializer.serialize, serializer.deserialize)

    @classmethod
    def noop(cls) -> Serializer[T, T]:
        return NoopSerializer()

    @classmethod
    def model[_T, _D](cls, model: type[JsonSerializable[_T, _D]]) -> Serializer[_T, _D]:
        return ModelSerializer(model)  # type: ignore

    @classmethod
    def json(cls) -> Serializer[T, bytes]:
        return JsonSerializer()

    def to_json(self) -> Serializer[T, bytes]:
        return self.pipe(JsonSerializer())

    def to_array(self) -> Serializer[list[T], list[D]]:
        return ArraySerializer(self)

    def to_map(self) -> Serializer[Mapping[str, T], Mapping[str, D]]:
        return MapSerializer(self)

    def pipe[E](self, other: Serializable[D, E]) -> Serializer[T, E]:
        return PipeSerializer(self, other)


class NoopSerializer[T](Serializer[T, T]):
    def __init__(self):
        super().__init__(lambda item: item, lambda item: item)

    def __repr__(self) -> str:
        return "NoopSerializer()"


class ModelSerializer[M: JsonSerializable, D](Serializer[M, D]):
    def __init__(self, model: type[JsonSerializable[M, D]]):
        self._model = model
        super().__init__(
            lambda item: item.to_json(), lambda item: model.from_json(item)
        )

    def __repr__(self) -> str:
        return f"ModelSerializer({self._model})"


class JsonSerializer[T](Serializer[T, bytes]):
    def __init__(self):
        super().__init__(
            self._serialize,
            self._deserialize,
        )

    def _serialize(self, item: T) -> bytes:
        return json.dumps(item).encode("utf-8")

    def _deserialize(self, item: bytes) -> T:
        decoded = item.decode("utf-8")
        try:
            return json.loads(decoded)
        except json.JSONDecodeError as e:
            raise SerializeError(f"Failed to deserialize JSON: {decoded}") from e

    def __repr__(self) -> str:
        return "JsonSerializer()"


class ArraySerializer[T, D](Serializer[list[T], list[D]]):
    def __init__(self, serializer: Serializable[T, D]):
        self._serializer = serializer
        super().__init__(
            lambda items: [serializer.serialize(item) for item in items],
            lambda items: [serializer.deserialize(item) for item in items],
        )

    def __repr__(self) -> str:
        return f"ArraySerializer({self._serializer})"


class MapSerializer[T, D](Serializer[Mapping[str, T], Mapping[str, D]]):
    def __init__(self, serializer: Serializable[T, D]):
        self._serializer = serializer
        super().__init__(
            self._serialize,
            self._deserialize,
        )

    def _serialize(self, items: Mapping[str, T]) -> Mapping[str, D]:
        return {key: self._serializer.serialize(value) for key, value in items.items()}

    def _deserialize(self, items: Mapping[str, D]) -> Mapping[str, T]:
        return {
            key: self._serializer.deserialize(value) for key, value in items.items()
        }

    def __repr__(self) -> str:
        return f"MapSerializer({self._serializer})"


class PipeSerializer[T, D, E](Serializer[T, E]):
    def __init__(self, a: Serializable[T, D], b: Serializable[D, E]):
        self._a = a
        self._b = b
        super().__init__(
            lambda item: b.serialize(a.serialize(item)),
            lambda item: a.deserialize(b.deserialize(item)),
        )

    def __repr__(self) -> str:
        return f"PipeSerializer({self._a}, {self._b})"



from .signal import Signal, SignalPermissions, SignalType
from .signal_extension import SIGNAL_EXTENSION_TYPE, SignalExtension

__all__ = [
    "Signal",
    "SignalPermissions",
    "SignalType",
    "SIGNAL_EXTENSION_TYPE",
    "SignalExtension",
]



from .logger_extension import (
    LOGGER_EXTENSION_TYPE,
    LOGGER_LOG_PERMISSION_ID,
    LOGGER_SERVER_ID,
    LoggerExtension,
)

__all__ = [
    "LOGGER_EXTENSION_TYPE",
    "LOGGER_LOG_PERMISSION_ID",
    "LOGGER_SERVER_ID",
    "LoggerExtension",
]



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TypedDict

from omu.app import App
from omu.bytebuffer import ByteReader, ByteWriter
from omu.extension.permission import PermissionType
from omu.extension.plugin.package_info import PackageInfo


@dataclass(frozen=True, slots=True)
class PermissionRequestPacket:
    request_id: str
    app: App
    permissions: list[PermissionType]

    @classmethod
    def serialize(cls, item: PermissionRequestPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.request_id)
        writer.write_string(json.dumps(item.app.to_json()))
        writer.write_string(
            json.dumps(tuple(map(PermissionType.to_json, item.permissions)))
        )
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> PermissionRequestPacket:
        with ByteReader(item) as reader:
            request_id = reader.read_string()
            app = App.from_json(json.loads(reader.read_string()))
            permissions = map(
                PermissionType.from_json, json.loads(reader.read_string())
            )
            return cls(request_id, app, list(permissions))


@dataclass(frozen=True, slots=True)
class PluginRequestPacket:
    request_id: str
    app: App
    packages: list[PackageInfo]

    @classmethod
    def serialize(cls, item: PluginRequestPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.request_id)
        writer.write_string(json.dumps(item.app.to_json()))
        writer.write_string(json.dumps(item.packages))
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> PluginRequestPacket:
        with ByteReader(item) as reader:
            request_id = reader.read_string()
            app = App.from_json(json.loads(reader.read_string()))
            plugins = map(PackageInfo, json.loads(reader.read_string()))
            return cls(request_id, app, list(plugins))


@dataclass(frozen=True, slots=True)
class AppInstallRequestPacket:
    request_id: str
    app: App

    @classmethod
    def serialize(cls, item: AppInstallRequestPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.request_id)
        writer.write_string(json.dumps(item.app.to_json()))
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> AppInstallRequestPacket:
        with ByteReader(item) as reader:
            request_id = reader.read_string()
            app = App.from_json(json.loads(reader.read_string()))
            return cls(request_id, app)


@dataclass(frozen=True, slots=True)
class AppUpdateRequestPacket:
    request_id: str
    old_app: App
    new_app: App

    @classmethod
    def serialize(cls, item: AppUpdateRequestPacket) -> bytes:
        writer = ByteWriter()
        writer.write_string(item.request_id)
        writer.write_string(json.dumps(item.old_app.to_json()))
        writer.write_string(json.dumps(item.new_app.to_json()))
        return writer.finish()

    @classmethod
    def deserialize(cls, item: bytes) -> AppUpdateRequestPacket:
        with ByteReader(item) as reader:
            request_id = reader.read_string()
            old_app = App.from_json(json.loads(reader.read_string()))
            new_app = App.from_json(json.loads(reader.read_string()))
            return cls(request_id, old_app, new_app)


class AppInstallResponse(TypedDict):
    accepted: bool


class AppUpdateResponse(TypedDict):
    accepted: bool



from __future__ import annotations

from dataclasses import dataclass, field

from omu.app import App
from omu.client import Client
from omu.extension import Extension, ExtensionType
from omu.extension.endpoint import EndpointType
from omu.extension.registry import RegistryPermissions, RegistryType
from omu.extension.table import TablePermissions, TableType
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.serializer import Serializer

SERVER_EXTENSION_TYPE = ExtensionType(
    "server", lambda client: ServerExtension(client), lambda: []
)

SERVER_APPS_READ_PERMISSION_ID = SERVER_EXTENSION_TYPE / "apps" / "read"
SERVER_APP_TABLE_TYPE = TableType.create_model(
    SERVER_EXTENSION_TYPE,
    "apps",
    App,
    permissions=TablePermissions(
        read=SERVER_APPS_READ_PERMISSION_ID,
    ),
)
SERVER_SESSIONS_READ_PERMISSION_ID = SERVER_EXTENSION_TYPE / "sessions" / "read"
SERVER_SESSION_TABLE_TYPE = TableType.create_model(
    SERVER_EXTENSION_TYPE,
    "sessions",
    App,
    permissions=TablePermissions(
        read=SERVER_SESSIONS_READ_PERMISSION_ID,
    ),
)
SERVER_SHUTDOWN_PERMISSION_ID = SERVER_EXTENSION_TYPE / "shutdown"
SHUTDOWN_ENDPOINT_TYPE = EndpointType[bool, bool].create_json(
    SERVER_EXTENSION_TYPE,
    "shutdown",
    permission_id=SERVER_SHUTDOWN_PERMISSION_ID,
)
REQUIRE_APPS_PACKET_TYPE = PacketType[list[Identifier]].create_json(
    SERVER_EXTENSION_TYPE,
    "require_apps",
    serializer=Serializer.model(Identifier).to_array(),
)
TRUSTED_ORIGINS_GET_PERMISSION_ID = SERVER_EXTENSION_TYPE / "trusted_origins" / "get"
TRUSTED_ORIGINS_SET_PERMISSION_ID = SERVER_EXTENSION_TYPE / "trusted_origins" / "set"
TRUSTED_ORIGINS_REGISTRY_TYPE = RegistryType[list[str]].create_json(
    SERVER_EXTENSION_TYPE,
    "trusted_origins",
    default_value=[],
    permissions=RegistryPermissions(
        read=TRUSTED_ORIGINS_GET_PERMISSION_ID,
        write=TRUSTED_ORIGINS_SET_PERMISSION_ID,
    ),
)
SESSION_OBSERVE_PACKET_TYPE = PacketType[list[Identifier]].create_json(
    SERVER_EXTENSION_TYPE,
    "session_observe",
    serializer=Serializer.model(Identifier).to_array(),
)
SESSION_CONNECT_PACKET_TYPE = PacketType[App].create_json(
    SERVER_EXTENSION_TYPE,
    "session_connect",
    serializer=Serializer.model(App),
)
SESSION_DISCONNECT_PACKET_TYPE = PacketType[App].create_json(
    SERVER_EXTENSION_TYPE,
    "session_disconnect",
    serializer=Serializer.model(App),
)


class ServerExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return SERVER_EXTENSION_TYPE

    def __init__(self, client: Client) -> None:
        self._client = client
        self.apps = client.tables.get(SERVER_APP_TABLE_TYPE)
        self.sessions = client.tables.get(SERVER_APP_TABLE_TYPE)
        self.required_apps: set[Identifier] = set()
        self.trusted_origins = client.registries.get(TRUSTED_ORIGINS_REGISTRY_TYPE)
        self.session_observers: dict[Identifier, SessionObserver] = {}
        client.network.register_packet(
            REQUIRE_APPS_PACKET_TYPE,
            SESSION_OBSERVE_PACKET_TYPE,
            SESSION_CONNECT_PACKET_TYPE,
            SESSION_DISCONNECT_PACKET_TYPE,
        )
        client.network.add_packet_handler(
            SESSION_CONNECT_PACKET_TYPE, self.handle_session_connect
        )
        client.network.add_packet_handler(
            SESSION_DISCONNECT_PACKET_TYPE, self.handle_session_disconnect
        )
        client.network.add_task(self.on_task)
        client.on_ready(self.on_ready)

    async def on_task(self) -> None:
        if self.required_apps:
            await self._client.send(REQUIRE_APPS_PACKET_TYPE, [*self.required_apps])

    async def on_ready(self) -> None:
        if self.session_observers:
            await self._client.send(
                SESSION_OBSERVE_PACKET_TYPE, [*self.session_observers]
            )

    async def shutdown(self, restart: bool = False) -> bool:
        return await self._client.endpoints.call(SHUTDOWN_ENDPOINT_TYPE, restart)

    def observe_session(
        self,
        app_id: Identifier,
        on_connect: Coro[[App], None] | None = None,
        on_disconnect: Coro[[App], None] | None = None,
    ) -> SessionObserver:
        if self._client.running:
            raise RuntimeError("Cannot require apps after the client has started")
        observer = self.session_observers.get(app_id) or SessionObserver()
        if on_connect:
            observer.on_connect(on_connect)
        if on_disconnect:
            observer.on_disconnect(on_disconnect)
        self.session_observers[app_id] = observer
        return observer

    async def handle_session_connect(self, app: App) -> None:
        observer = self.session_observers.get(app.id)
        if observer is None:
            return
        for callback in observer.on_connect_callbacks:
            await callback(app)

    async def handle_session_disconnect(self, app: App) -> None:
        observer = self.session_observers.get(app.id)
        if observer is None:
            return
        for callback in observer.on_disconnect_callbacks:
            await callback(app)

    def require(self, *app_ids: Identifier) -> None:
        if self._client.running:
            raise RuntimeError("Cannot require apps after the client has started")
        self.required_apps.update(app_ids)


@dataclass(frozen=True, slots=True)
class SessionObserver:
    on_connect_callbacks: list[Coro[[App], None]] = field(default_factory=list)
    on_disconnect_callbacks: list[Coro[[App], None]] = field(default_factory=list)

    def on_connect(self, coro: Coro[[App], None]) -> Coro[[App], None]:
        self.on_connect_callbacks.append(coro)
        return coro

    def on_disconnect(self, coro: Coro[[App], None]) -> Coro[[App], None]:
        self.on_disconnect_callbacks.append(coro)
        return coro



from __future__ import annotations

import abc
from collections.abc import AsyncGenerator, Callable, Mapping
from dataclasses import dataclass
from typing import (
    NotRequired,
    TypedDict,
)

from omu.event_emitter import EventEmitter, Unlisten
from omu.helper import AsyncCallback, Coro
from omu.identifier import Identifier
from omu.interface import Keyable
from omu.serializer import JsonSerializable, Serializable, Serializer


class TableConfig(TypedDict):
    cache_size: NotRequired[int]


class Table[T](abc.ABC):
    @property
    @abc.abstractmethod
    def cache(self) -> Mapping[str, T]: ...

    @abc.abstractmethod
    def set_cache_size(self, size: int) -> None: ...

    @abc.abstractmethod
    async def get(self, key: str) -> T | None: ...

    @abc.abstractmethod
    async def get_many(self, *keys: str) -> dict[str, T]: ...

    @abc.abstractmethod
    async def add(self, *items: T) -> None: ...

    @abc.abstractmethod
    async def update(self, *items: T) -> None: ...

    @abc.abstractmethod
    async def remove(self, *items: T) -> None: ...

    @abc.abstractmethod
    async def has(self, key: str) -> bool: ...

    @abc.abstractmethod
    async def has_many(self, *keys: str) -> dict[str, bool]: ...

    @abc.abstractmethod
    async def has_all(self, *keys: str) -> bool: ...

    @abc.abstractmethod
    async def has_any(self, *keys: str) -> bool: ...

    @abc.abstractmethod
    async def clear(self) -> None: ...

    @abc.abstractmethod
    async def fetch_items(
        self,
        limit: int,
        backward: bool = False,
        cursor: str | None = None,
    ) -> Mapping[str, T]: ...

    @abc.abstractmethod
    async def fetch_range(self, start: str, end: str) -> dict[str, T]: ...

    @abc.abstractmethod
    async def fetch_all(self) -> dict[str, T]: ...

    @abc.abstractmethod
    async def iterate(
        self,
        backward: bool = False,
        cursor: str | None = None,
    ) -> AsyncGenerator[T, None]: ...

    @abc.abstractmethod
    async def size(self) -> int: ...

    @abc.abstractmethod
    def listen(
        self, listener: AsyncCallback[Mapping[str, T]] | None = None
    ) -> Unlisten: ...

    @abc.abstractmethod
    def proxy(self, callback: Coro[[T], T | None]) -> Unlisten: ...

    @abc.abstractmethod
    def set_config(self, config: TableConfig) -> None: ...

    @property
    @abc.abstractmethod
    def event(self) -> TableEvents[T]: ...


class TableEvents[T]:
    def __init__(
        self,
        table: Table[T],
    ) -> None:
        self.unlisten: Unlisten | None = None

        def listen():
            self.unlisten = table.listen()

        def unlisten():
            if self.unlisten:
                self.unlisten()

        self.add: EventEmitter[Mapping[str, T]] = EventEmitter(
            on_subscribe=listen, on_empty=unlisten
        )
        self.update: EventEmitter[Mapping[str, T]] = EventEmitter(
            on_subscribe=listen, on_empty=unlisten
        )
        self.remove: EventEmitter[Mapping[str, T]] = EventEmitter(
            on_subscribe=listen, on_empty=unlisten
        )
        self.clear: EventEmitter[[]] = EventEmitter(
            on_subscribe=listen, on_empty=unlisten
        )
        self.cache_update: EventEmitter[Mapping[str, T]] = EventEmitter(
            on_subscribe=listen, on_empty=unlisten
        )


type ModelEntry[T: Keyable, D] = JsonSerializable[T, D]


@dataclass(frozen=True, slots=True)
class TablePermissions:
    all: Identifier | None = None
    read: Identifier | None = None
    write: Identifier | None = None
    remove: Identifier | None = None
    proxy: Identifier | None = None


@dataclass(frozen=True, slots=True)
class TableType[T]:
    id: Identifier
    serializer: Serializable[T, bytes]
    key_function: Callable[[T], str]
    permissions: TablePermissions | None = None

    @classmethod
    def create_model[_T: Keyable, _D](
        cls,
        identifier: Identifier,
        name: str,
        model_type: type[ModelEntry[_T, _D]],
        permissions: TablePermissions | None = None,
    ) -> TableType[_T]:
        return TableType(
            id=identifier / name,
            serializer=Serializer.model(model_type).to_json(),
            key_function=lambda item: item.key(),
            permissions=permissions,
        )

    @classmethod
    def create_json[_T](
        cls,
        identifier: Identifier,
        name: str,
        key: Callable[[_T], str],
        permissions: TablePermissions | None = None,
    ) -> TableType[_T]:
        return TableType(
            id=identifier / name,
            serializer=Serializer.json(),
            key_function=key,
            permissions=permissions,
        )



from __future__ import annotations

import abc
from collections.abc import Callable

from omu.client import Client
from omu.identifier import Identifier


class Extension(abc.ABC):
    @property
    @abc.abstractmethod
    def type(self) -> ExtensionType: ...


class ExtensionType[T: Extension](Identifier):
    name: str
    create: Callable[[Client], T]
    dependencies: Callable[[], list[ExtensionType]]

    def __init__(
        self,
        name: str,
        create: Callable[[Client], T],
        dependencies: Callable[[], list[ExtensionType]],
    ) -> None:
        super().__init__("ext", name)
        self.name = name
        self.create = create
        self.dependencies = dependencies

    def key(self) -> str:
        return self.name
